// Copyright 2003 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

// In addition, as a special exception, the Regents of the University of
// California give permission to link the code of this program with libraries
// that provide specific optimized fast Fourier transform (FFT) functions and
// distribute a linked executable.  You must obey the GNU General Public 
// License in all respects for all of the code used other than the FFT library
// itself.  Any modification required to support these libraries must be
// distributed in source code form.  If you modify this file, you may extend 
// this exception to your version of the file, but you are not obligated to 
// do so. If you do not wish to do so, delete this exception statement from 
// your version.

// gaussfit.C
// $Id: gaussfit.cpp,v 1.21.2.9 2006/04/12 20:53:16 korpela Exp $
//

// The purpose of gaussian fitting is to find signals that rise and
// fall in power over time at a rate consistent with the beam
// pattern of the telescope.

#include "config.h"

#include <stdio.h>
#include <math.h>

#include <algorithm>

// debug stuff
#define DEBUG_POT
//#define DUMP_POWER_SPECTRA
//#define DUMP_GAUSSIAN
#define POT_TO_DUMP 32
#define TOFF_TO_DUMP 15
#define FFT_TO_DUMP 131072
int gul_PoT;
int gul_Fftl;
// end debug stuff

#ifdef BOINC_APP_GRAPHICS
#include "graphics_api.h"
#ifdef DYNAMIC_GRAPHICS
#include "graphics_lib.h"
#endif
#endif

#ifdef __SSE3__
#pragma message("---- SSE3 ----")
#include <pmmintrin.h>
#elif defined __SSE2__
#pragma message("---- SSE2 ----")
#include <emmintrin.h>
#elif defined __SSE__
#pragma message("---- SSE ----")
#include <xmmintrin.h>
#endif

#ifdef USE_SSE2
#pragma message("------> Enabling SSE3 emulation by SSE2 ----")
#include "SSE3_emulate.h"
#endif

//#include "util.h"
#include "s_util.h"
#include "analyze.h"
#include "seti.h"
#include "worker.h"
#include "analyzeFuncs.h"
#include "analyzeReport.h"
#include "analyzePoT.h"
#include "lcgamm.h"
#include "gaussfit.h"
#include "chirpfft.h"

#ifdef BOINC_APP_GRAPHICS
#include "sah_gfx.h"
#endif

#if defined(USE_IPP)
  #ifdef _WIN64
	#pragma message ("----x64 IPP includes----")
	#include "ipp_x64_static.h"
  #else
	#include "ipp_includes.h"
  #endif	
#endif // USE_IPP

//#include <Accelerate/Accelerate.h>

#if defined(__VEC__) || defined(__SSE3__)
#define ZERO		(_mm_set_ps(0.0f, 0.0f, 0.0f, 0.0f))
#define ONE         (_mm_set_ps(1.0f, 1.0f, 1.0f, 1.0f))
#define TWO         (_mm_set_ps(2.0f, 2.0f, 2.0f, 2.0f))
#define RECIP_TWO	(_mm_set_ps(1.0f/2.0f, 1.0f/2.0f, 1.0f/2.0f, 1.0f/2.0f))
#define RECIP_THREE	(_mm_set_ps(1.0f/3.0f, 1.0f/3.0f, 1.0f/3.0f, 1.0f/3.0f))
#define RECIP_FOUR	(_mm_set_ps(1.0f/4.0f, 1.0f/4.0f, 1.0f/4.0f, 1.0f/4.0f))
#define RECIP_FIVE	(_mm_set_ps(1.0f/5.0f, 1.0f/5.0f, 1.0f/5.0f, 1.0f/5.0f))
#endif

#ifdef USE_PPC_OPTIMIZATIONS
#define USE_VECTORIZED_GAUSSIAN
#elif defined(USE_I386_OPTIMIZATIONS)
 #define USE_VECTORIZED_GAUSSIAN //Jason: vectorized gaussian using dodgy IPP conv() method/routine
#endif

bool dump_pot = false;
//float gauss_display_power_thresh = 0;

float f_GetPeak(
  float fp_PoT[],
  int ul_TOffset,
  int ul_HalfSumLength,
  float f_MeanPower,
  float f_PeakScaleFactor,
  float f_weight[]
) {
  // Peak power is calculated as the weighted
  // sum of all powers within ul_HalfSumLength
  // of the assumed gaussian peak.
  // The weights are given by the gaussian function itself.
  // BUG WATCH : for the f_PeakScaleFactor to work,
  // ul_HalfSumLength *must* be set to sigma.

  int i;
  int vweightCenter = PoTInfo.GaussTOffsetStop-1;
  float f_sum;

  f_sum = 0.0;

  // Find a weighted sum
  for (i = ul_TOffset - ul_HalfSumLength; i <= ul_TOffset + ul_HalfSumLength; i++) {
//    f_sum += (fp_PoT[i] - f_MeanPower) * f_weight[abs(i-ul_TOffset)];
    f_sum += (fp_PoT[i] - f_MeanPower) * f_weight[vweightCenter+(i-ul_TOffset)];
  }
  
  analysis_state.FLOP_counter+=6.0*ul_HalfSumLength;
  return(f_sum * f_PeakScaleFactor);
}


float f_GetChiSq(
  float fp_PoT[],
  int ul_PowerLen,
  int ul_TOffset,
  float f_PeakPower,
  float f_MeanPower,
  float f_weight[],
  float *xsq_null=0
) {
  // We calculate our assumed gaussian powers
  // on the fly as we try to fit them to the
  // actual powers at each point along the PoT.

  int i;
  float f_ChiSq=0,f_null_hyp=0, f_PredictedPower, f_tot_weight=0;
  double rebin=swi.nsamples/ChirpFftPairs[analysis_state.icfft].FftLen
                  /ul_PowerLen;

  for (i = 0; i < ul_PowerLen; i++) {
    f_PredictedPower = f_MeanPower +
                       f_PeakPower * f_weight[abs(i-ul_TOffset)] ;
    f_tot_weight+=f_weight[abs(i-ul_TOffset)];
#ifdef DUMP_GAUSSIAN
    if ((gul_PoT == POT_TO_DUMP && ul_TOffset == TOFF_TO_DUMP) &&
        gul_Fftl == FFT_TO_DUMP) {
      fprintf(stdout, "%f\n", f_PredictedPower);
    }
#endif
    // ChiSq in this realm is:
    //  sum[0:i]( (observed power - expected power)^2 / expected variance )
    // The power of a signal is:
    //  power = (noise + signal)^2 = noise^2 + signal^2 + 2*noise*signal
    // With mean power normalization, noise becomes 1, leaving:
    //  power = signal^2 +or- 2*signal + 1
    f_PredictedPower/=f_MeanPower;
    double signal=f_PredictedPower;
    double noise=(2.0*sqrt(std::max(f_PredictedPower,1.0f)-1)+1);

    f_ChiSq += 
      (static_cast<float>(rebin)*SQUARE(fp_PoT[i]/f_MeanPower - signal)/noise);
    f_null_hyp+=
      (static_cast<float>(rebin)*SQUARE(fp_PoT[i]/f_MeanPower-1)/noise);
  }

  analysis_state.FLOP_counter+=20.0*ul_PowerLen+5;
  f_ChiSq/=ul_PowerLen;
  f_null_hyp/=ul_PowerLen;

#ifdef DUMP_GAUSSIAN
  if (gul_PoT == POT_TO_DUMP
      && ul_TOffset == TOFF_TO_DUMP
      && gul_Fftl == FFT_TO_DUMP
     ) {
    for (i = 0; i < ul_PowerLen; i++) {
      fprintf(stdout, "%f\n", fp_PoT[i]);
    }
  }
#endif
  if (xsq_null) *xsq_null=f_null_hyp;
  return f_ChiSq;
}


#ifdef __VEC__
inline vector float vec_fpsplat(float f_Constant) {
  vector unsigned char vuc_Splat;
  vector float vf_Constant;
  
  vuc_Splat = vec_lvsl(0, &f_Constant);
  vf_Constant = vec_lde(0, &f_Constant);
  vuc_Splat = (vector unsigned char) vec_splat((vector float) vuc_Splat, 0);
  vf_Constant = vec_perm(vf_Constant, vf_Constant, vuc_Splat);
  
  return vf_Constant;
}


inline vector float vec_rsqrt(vector float v) {
  const vector float _0 = (vector float)(-.0); 
  // obtain estimate
  vector float y = vec_rsqrte(v);
  // one round of Newton-Raphson
  y = vec_madd(vec_nmsub(v,vec_madd(y,y,_0),(vector float)(1.0)),vec_madd(y,(vector float)(0.5),_0),y);
  
  return y;
}


inline vector float vec_sqrt(vector float v) {
  const vector float _0 = (vector float)(-.0);
  return vec_sel(vec_madd(v,vec_rsqrt(v),_0),_0,vec_cmpeq(v,_0));
}


inline vector float vec_recip(vector float v) {
  // obtain estimate
  vector float estimate = vec_re( v );
  // one round of Newton-Raphson
  return vec_madd( vec_nmsub( estimate, v, (vector float) (1.0) ), estimate, estimate );
}  


float v_GetChiSq(
  float fp_PoT[],
  int ul_PowerLen,
  int ul_TOffset,
  float f_PeakPower,
  float f_MeanPower,
  float f_weight[],
  float *xsq_null=0
) {
  // We calculate our assumed gaussian powers
  // on the fly as we try to fit them to the
  // actual powers at each point along the PoT.
  
  int i;
  int vEnd, vul_PowerLen;
  float f_ChiSq=0,f_null_hyp=0, f_PredictedPower;
  vector float v_PeakPower, v_MeanPower, v_rMeanPower, v_rebin;
  vector float v_TempChiSq = ZERO, v_ChiSq, v_temp_null_hyp = ZERO, v_null_hyp, v_PredictedPower, v_weight;
  double rebin=swi.nsamples/ChirpFftPairs[analysis_state.icfft].FftLen
    /ul_PowerLen;
  
  // splat Altivec versions of function constants
  v_PeakPower = vec_fpsplat(f_PeakPower);
  v_MeanPower = vec_fpsplat(f_MeanPower);
  v_rMeanPower = vec_fpsplat(1.0f / f_MeanPower);
  v_rebin = vec_fpsplat((float) rebin);
  
  // set up unaligned loads for weight array
  int offwt = PoTInfo.GaussTOffsetStop - 1 - ul_TOffset;
  vector float msqwt = vec_ld(0, f_weight + offwt);
  vector unsigned char maskwt = vec_add(vec_lvsl(-1L, f_weight + offwt), vec_splat_u8(1));
  
  vul_PowerLen = ul_PowerLen - (ul_PowerLen & 0x3);
  for (i = 0; i < vul_PowerLen; i += 4) {
    const float *p = fp_PoT + i;
    const float *wt = f_weight + offwt + i;
    vector float lsqwt = vec_ld(15, wt);
    vector float v_PoT = vec_ld(0, p);
    vector float v_weight = vec_perm(msqwt, lsqwt, maskwt);
    vector float v_noise, v_rnoise;

    v_PredictedPower = vec_madd(v_PeakPower, v_weight, v_MeanPower);
    v_PredictedPower = vec_madd(v_PredictedPower, v_rMeanPower, ZERO);
    v_noise = vec_madd(vec_sqrt(vec_sub(vec_max(v_PredictedPower, ONE), ONE)), TWO, ONE);
    v_rnoise = vec_recip(v_noise);
    
    v_TempChiSq = vec_madd(vec_madd(vec_madd(vec_nmsub(v_PoT, v_rMeanPower, v_PredictedPower),
                                         vec_nmsub(v_PoT, v_rMeanPower, v_PredictedPower),
                                         ZERO),
                                v_rnoise,
                                ZERO),
                       v_rebin,
                       v_TempChiSq);
    v_temp_null_hyp = vec_madd(vec_madd(vec_madd(vec_nmsub(v_PoT, v_rMeanPower, ONE),
                                            vec_nmsub(v_PoT, v_rMeanPower, ONE),
                                            ZERO),
                                   v_rnoise,
                                   ZERO),
                          v_rebin,
                          v_temp_null_hyp);
    
    msqwt = lsqwt;
  }
  v_ChiSq = v_TempChiSq;
  v_null_hyp = v_temp_null_hyp;
  
  // handle tail elements
  for (; i < ul_PowerLen; i++) {
    f_PredictedPower = f_MeanPower +
    f_PeakPower * f_weight[offwt+i] ;
#ifdef DUMP_GAUSSIAN
    if ((gul_PoT == POT_TO_DUMP && ul_TOffset == TOFF_TO_DUMP) &&
        gul_Fftl == FFT_TO_DUMP) {
      fprintf(stdout, "%f\n", f_PredictedPower);
    }
#endif
    // ChiSq in this realm is:
    //  sum[0:i]( (observed power - expected power)^2 / expected variance )
    // The power of a signal is:
    //  power = (noise + signal)^2 = noise^2 + signal^2 + 2*noise*signal
    // With mean power normalization, noise becomes 1, leaving:
    //  power = signal^2 +or- 2*signal + 1
    f_PredictedPower/=f_MeanPower;
    double signal=f_PredictedPower;
    double noise=(2.0*sqrt(std::max(f_PredictedPower,1.0f)-1)+1);
    
    f_ChiSq += 
      (static_cast<float>(rebin)*SQUARE(fp_PoT[i]/f_MeanPower - signal)/noise);
    f_null_hyp+=
      (static_cast<float>(rebin)*SQUARE(fp_PoT[i]/f_MeanPower-1)/noise);
  }
  
  // sum vector and scalar elements
  float *vp_ChiSq = (float *) &v_ChiSq;
  float *vp_null_hyp = (float *) &v_null_hyp;
  f_ChiSq += (vp_ChiSq[0] + vp_ChiSq[1]) + (vp_ChiSq[2] + vp_ChiSq[3]);
  f_null_hyp += (vp_null_hyp[0] + vp_null_hyp[1]) + (vp_null_hyp[2] + vp_null_hyp[3]);
  
  analysis_state.FLOP_counter+=20.0*ul_PowerLen+5;
  f_ChiSq/=ul_PowerLen;
  f_null_hyp/=ul_PowerLen;
  
#ifdef DUMP_GAUSSIAN
  if (gul_PoT == POT_TO_DUMP
      && ul_TOffset == TOFF_TO_DUMP
      && gul_Fftl == FFT_TO_DUMP
      ) {
    for (i = 0; i < ul_PowerLen; i++) {
      fprintf(stdout, "%f\n", fp_PoT[i]);
    }
  }
#endif
  if (xsq_null) *xsq_null=f_null_hyp;
  return f_ChiSq;
}


#elif defined(__SSE3__)
inline vFloat __attribute__((__always_inline__)) vec_rsqrt(vFloat v) {
  // obtain estimate
  vFloat y = _mm_rsqrt_ps(v);
  // one round of Newton-Raphson
  y = _mm_add_ps(_mm_mul_ps(_mm_sub_ps(ONE, _mm_mul_ps(v, _mm_mul_ps(y, y))), _mm_mul_ps(y, RECIP_TWO)), y);
  
  return y;
}


inline vFloat __attribute__((__always_inline__)) vec_sqrt(vFloat v) {
  const vFloat _0 = _mm_set1_ps(-.0);
  vFloat cmp = _mm_cmpeq_ps(v, _0);
  
  return _mm_andnot_ps(cmp, _mm_mul_ps(v, vec_rsqrt(v)));
}


inline vFloat __attribute__((__always_inline__)) vec_recip(vFloat v) {
  // obtain estimate
  vFloat estimate = _mm_rcp_ps( v );
  // one round of Newton-Raphson
  return _mm_add_ps(_mm_mul_ps(_mm_sub_ps(ONE, _mm_mul_ps(estimate, v)), estimate), estimate);
}  


float v_GetChiSq(
                 float fp_PoT[],
                 int ul_PowerLen,
                 int ul_TOffset,
                 float f_PeakPower,
                 float f_MeanPower,
                 float f_weight[],
                 float *xsq_null=0
                 ) {
  // We calculate our assumed gaussian powers
  // on the fly as we try to fit them to the
  // actual powers at each point along the PoT.
  
  int i;
  int vEnd, vul_PowerLen;
  float f_ChiSq=0,f_null_hyp=0, f_PredictedPower;
  vFloat v_PeakPower, v_MeanPower, v_rMeanPower, v_rebin;
  vFloat v_ChiSq = _mm_setzero_ps(), v_null_hyp = _mm_setzero_ps(), v_PredictedPower, v_weight;
  double rebin=swi.nsamples/ChirpFftPairs[analysis_state.icfft].FftLen
    /ul_PowerLen;
  
  // splat SSE versions of function constants
  v_PeakPower = _mm_set1_ps(f_PeakPower);
  v_MeanPower = _mm_set1_ps(f_MeanPower);
  v_rMeanPower = _mm_set1_ps(1.0f / f_MeanPower);
  v_rebin = _mm_set1_ps((float) rebin);
  
  // set up unaligned loads for weight array
  int offwt = PoTInfo.GaussTOffsetStop - 1 - ul_TOffset;
  
  vul_PowerLen = ul_PowerLen - (ul_PowerLen & 0x3);
  for (i = 0; i < vul_PowerLen; i += 4) {
    const float *p = fp_PoT + i;
    const float *wt = f_weight + offwt + i;
    vFloat v_PoT = _mm_load_ps(p);
//    vFloat v_weight = _mm_loadu_ps(wt);
    vFloat v_weight = (vFloat) _mm_lddqu_si128((__m128i *) wt);
    vFloat v_noise, v_rnoise;
    
    v_PredictedPower = _mm_add_ps(_mm_mul_ps(v_PeakPower, v_weight), v_MeanPower);
    v_PredictedPower = _mm_mul_ps(v_PredictedPower, v_rMeanPower);

    v_noise = _mm_add_ps(_mm_mul_ps(vec_sqrt(_mm_sub_ps(_mm_max_ps(v_PredictedPower, ONE), ONE)), TWO), ONE);
    v_rnoise = vec_recip(v_noise);
    
    v_ChiSq = _mm_add_ps(_mm_mul_ps(_mm_mul_ps(_mm_mul_ps(_mm_sub_ps(v_PredictedPower, _mm_mul_ps(v_PoT, v_rMeanPower)),
                                                          _mm_sub_ps(v_PredictedPower, _mm_mul_ps(v_PoT, v_rMeanPower))),
                                               v_rnoise),
                                    v_rebin),
                         v_ChiSq);
    v_null_hyp = _mm_add_ps(_mm_mul_ps(_mm_mul_ps(_mm_mul_ps(_mm_sub_ps(ONE, _mm_mul_ps(v_PoT, v_rMeanPower)),
                                                             _mm_sub_ps(ONE, _mm_mul_ps(v_PoT, v_rMeanPower))),
                                                  v_rnoise),
                                       v_rebin),
                            v_null_hyp);
  }
//  v_ChiSq = _mm_hadd_ps(v_ChiSq, v_ChiSq);
//  v_ChiSq = _mm_hadd_ps(v_ChiSq, v_ChiSq);
  v_ChiSq = _mm_add_ps(v_ChiSq, _mm_movehl_ps(v_ChiSq, v_ChiSq));
  v_ChiSq = _mm_add_ss(v_ChiSq, _mm_movehdup_ps(v_ChiSq));
//  v_null_hyp = _mm_hadd_ps(v_null_hyp, v_null_hyp);
//  v_null_hyp = _mm_hadd_ps(v_null_hyp, v_null_hyp);  
  v_null_hyp = _mm_add_ps(v_null_hyp, _mm_movehl_ps(v_null_hyp, v_null_hyp));
  v_null_hyp = _mm_add_ss(v_null_hyp, _mm_movehdup_ps(v_null_hyp));
//  _mm_store_ss(&f_ChiSq, v_ChiSq);
//  _mm_store_ss(&f_null_hyp, v_null_hyp);
  f_ChiSq = _mm_cvtss_f32(v_ChiSq);
  f_null_hyp = _mm_cvtss_f32(v_null_hyp);
  
  // handle tail elements
  for (; i < ul_PowerLen; i++) {
    f_PredictedPower = f_MeanPower +
    f_PeakPower * f_weight[offwt+i] ;
#ifdef DUMP_GAUSSIAN
    if ((gul_PoT == POT_TO_DUMP && ul_TOffset == TOFF_TO_DUMP) &&
        gul_Fftl == FFT_TO_DUMP) {
      fprintf(stdout, "%f\n", f_PredictedPower);
    }
#endif
    // ChiSq in this realm is:
    //  sum[0:i]( (observed power - expected power)^2 / expected variance )
    // The power of a signal is:
    //  power = (noise + signal)^2 = noise^2 + signal^2 + 2*noise*signal
    // With mean power normalization, noise becomes 1, leaving:
    //  power = signal^2 +or- 2*signal + 1
    f_PredictedPower/=f_MeanPower;
    double signal=f_PredictedPower;
    double noise=(2.0*sqrt(std::max(f_PredictedPower,1.0f)-1)+1);
    
    f_ChiSq += 
      (static_cast<float>(rebin)*SQUARE(fp_PoT[i]/f_MeanPower - signal)/noise);
    f_null_hyp+=
      (static_cast<float>(rebin)*SQUARE(fp_PoT[i]/f_MeanPower-1)/noise);
  }
  
  analysis_state.FLOP_counter+=20.0*ul_PowerLen+5;
  f_ChiSq/=ul_PowerLen;
  f_null_hyp/=ul_PowerLen;
  
#ifdef DUMP_GAUSSIAN
  if (gul_PoT == POT_TO_DUMP
      && ul_TOffset == TOFF_TO_DUMP
      && gul_Fftl == FFT_TO_DUMP
      ) {
    for (i = 0; i < ul_PowerLen; i++) {
      fprintf(stdout, "%f\n", fp_PoT[i]);
    }
  }
#endif
  if (xsq_null) *xsq_null=f_null_hyp;
  return f_ChiSq;
}
#endif


float f_GetTrueMean(
  float fp_PoT[],
  int ul_PowerLen,
  float f_TotalPower,
  int ul_TOffset,
  int ul_ExcludeLen
) {
  // TrueMean is the mean power of the data set minus all power
  // out to ExcludeLen from our current TOffset.
  int i, i_start, i_lim;
  float f_ExcludePower = 0;

  // take care that we do not add to exclude power beyond PoT bounds!
  i_start = std::max(ul_TOffset - ul_ExcludeLen, 0);
  i_lim = std::min<int>(ul_TOffset + ul_ExcludeLen + 1, swi.analysis_cfg.gauss_pot_length);
  for (i = i_start; i < i_lim; i++) {
    f_ExcludePower += fp_PoT[i];
  }
  
  analysis_state.FLOP_counter+=(double)(i_lim-i_start+5);
  return((f_TotalPower - f_ExcludePower) / (ul_PowerLen - (i - i_start)));
}


#ifdef __VEC__
float v_GetTrueMean(
                    float fp_PoT[],
                    int ul_PowerLen,
                    float f_TotalPower,
                    int ul_TOffset,
                    int ul_ExcludeLen
                    ) {
  // TrueMean is the mean power of the data set minus all power
  // out to ExcludeLen from our current TOffset.
  int i, i_start, i_lim, vStart, vEnd;
  float f_ExcludePower = 0;
  vector float v_TempExcludePower = ZERO, v_ExcludePower;
  
  // take care that we do not add to exclude power beyond PoT bounds!
  i_start = std::max(ul_TOffset - ul_ExcludeLen, 0);
  i_lim = std::min<int>(ul_TOffset + ul_ExcludeLen + 1, swi.analysis_cfg.gauss_pot_length);
  
  vStart = i_start + 3;
  vStart = vStart - (vStart & 3);
  vEnd = i_lim - (i_lim & 3);
  
  // handle leading elements
  for (i = i_start; i < vStart; i++) {
    f_ExcludePower += fp_PoT[i];
  }
  for (; i < vEnd; i += 4) {
    v_TempExcludePower = vec_add(vec_ld(0, fp_PoT + i), v_TempExcludePower);
  }
  v_ExcludePower = v_TempExcludePower;
  // handle tail elements
  for (; i < i_lim; i++) {
    f_ExcludePower += fp_PoT[i];
  }
  
  // sum vector and scalar elements
  float *vp_ExcludePower = (float *) &v_ExcludePower;
  f_ExcludePower += (vp_ExcludePower[0] + vp_ExcludePower[1]) + (vp_ExcludePower[2] + vp_ExcludePower[3]);  
  
  analysis_state.FLOP_counter+=(double)(i_lim-i_start+5);
  return((f_TotalPower - f_ExcludePower) / (ul_PowerLen - (i - i_start)));
}


#elif defined(__SSE3__)
float v_GetTrueMean(
                    float fp_PoT[],
                    int ul_PowerLen,
                    float f_TotalPower,
                    int ul_TOffset,
                    int ul_ExcludeLen
                    ) {
  // TrueMean is the mean power of the data set minus all power
  // out to ExcludeLen from our current TOffset.
  int i, i_start, i_lim, vStart, vEnd;
  float vf_ExcludePower, f_ExcludePower = 0;
  vFloat v_ExcludePower = _mm_setzero_ps();

  // take care that we do not add to exclude power beyond PoT bounds!
  i_start = std::max(ul_TOffset - ul_ExcludeLen, 0);
  i_lim = std::min<int>(ul_TOffset + ul_ExcludeLen + 1, swi.analysis_cfg.gauss_pot_length);
  
  vStart = i_start + 3;
  vStart = vStart - (vStart & 3);
  vEnd = i_lim - (i_lim & 3);
  
  // handle leading elements
  for (i = i_start; i < vStart; i++) {
    f_ExcludePower += fp_PoT[i];
  }
  for (; i < vEnd; i += 4) {
    v_ExcludePower = _mm_add_ps(v_ExcludePower, _mm_load_ps(fp_PoT + i));
  }
//  v_ExcludePower = _mm_hadd_ps(v_ExcludePower, v_ExcludePower);
//  v_ExcludePower = _mm_hadd_ps(v_ExcludePower, v_ExcludePower);
  v_ExcludePower = _mm_add_ps(v_ExcludePower, _mm_movehl_ps(v_ExcludePower, v_ExcludePower));
  v_ExcludePower = _mm_add_ss(v_ExcludePower, _mm_movehdup_ps(v_ExcludePower));
//  _mm_store_ss(&vf_ExcludePower, v_ExcludePower);
  vf_ExcludePower = _mm_cvtss_f32(v_ExcludePower);
  // handle tail elements
  for (; i < i_lim; i++) {
    f_ExcludePower += fp_PoT[i];
  }
    
  f_ExcludePower += vf_ExcludePower;
  analysis_state.FLOP_counter+=(double)(i_lim-i_start+5);
  return((f_TotalPower - f_ExcludePower) / (ul_PowerLen - (i - i_start)));  
}
#endif


float f_GetPeakScaleFactor(float f_sigma) {
  // The PeakScaleFactor is calculated such that when used in f_GetPeak(),
  // the actual peak power can be extracted from a weighted sum.
  // This sum (see f_GetPeak()), is calculated as :
  // sum = SUM[x from -sigma to +sigma] of (gaussian weights * our data)
  // The gaussian weights are e^(-x^2 / sigma^2).
  // Our data is A(e^(-x^2 / sigma^2)), where 'A' is the peak power.
  // Through algebraic manipulation, we have:
  // A = sum * (1 / SUM[x from -sigma to +sigma] of (e^(-x^2 / sigma^2))^2.
  // The factor by which we multiply the sum is the PeakScaleFactor.
  // It is completely determined by sigma.

  int i, i_s = static_cast<int>(floor(f_sigma+0.5));
  float f_sigma_sq = f_sigma*f_sigma;
  float f_sum = 0.0;

  for (i = -i_s; i <= i_s; i++) {
    f_sum += static_cast<float>(EXP(i, 0, f_sigma_sq));
  }
 
  analysis_state.FLOP_counter+=(13.0*f_sigma+3);
  return(1 / f_sum);
}


int ChooseGaussEvent(
  int ifft,
  float PeakPower,
  float TrueMean,
  float ChiSq,
  float null_ChiSq,
  int bin,
  float sigma,
  float PoTMaxPower,
  float fp_PoT[]
) {

  float scale_factor;
  float gauss_bins, gauss_dof, null_dof;
#ifdef BOINC_APP_GRAPHICS
  BOOLEAN graphics = !nographics();
#endif
  BOOLEAN chisq = (ChiSq <= swi.analysis_cfg.gauss_chi_sq_thresh);
  double score;
  
  // precalculate some values, since values set in the workunit header won't change
  static float lcgf_gauss = lcgf(0.5*(swi.analysis_cfg.gauss_pot_length-2),
                                 swi.analysis_cfg.gauss_chi_sq_thresh*0.5*swi.analysis_cfg.gauss_pot_length);
  static float lcgf_null = lcgf(0.5*(swi.analysis_cfg.gauss_pot_length-1),
                                swi.analysis_cfg.gauss_null_chi_sq_thresh*0.5*swi.analysis_cfg.gauss_pot_length);

  // gaussian info
#define INIT_GAUSSIAN                                                                                        \
  gi = new(GAUSS_INFO);                                                                                      \
  gi->bin                  = bin;                                                                            \
  gi->fft_ind              = ifft;                                                                           \
  gi->g.chirp_rate 	  = ChirpFftPairs[analysis_state.icfft].ChirpRate;                                       \
  gi->g.fft_len    	  = ChirpFftPairs[analysis_state.icfft].FftLen;                                          \
  gi->g.sigma      	  = sigma;                                                                               \
  gi->g.peak_power 	  = PeakPower;                                                                           \
  gi->g.mean_power 	  = TrueMean;                                                                            \
  gi->g.chisqr     	  = ChiSq;                                                                               \
  gi->g.null_chisqr     	  = null_ChiSq;                                                                     \
  gi->g.freq       	  = cnvt_bin_hz(bin,  gi->g.fft_len);                                                    \
  double t_offset=(((double)gi->fft_ind+0.5)/swi.analysis_cfg.gauss_pot_length)*                             \
                  PoTInfo.WUDuration;                                                                        \
  gi->g.detection_freq=calc_detection_freq(gi->g.freq,gi->g.chirp_rate,t_offset);                            \
  gi->g.time       	  = swi.time_recorded+t_offset/86400.0;                                                  \
  gi->g.max_power  = PoTMaxPower;                                                                            \
  gi->score        = score;                                                                                  \
  time_to_ra_dec(gi->g.time, &gi->g.ra, &gi->g.decl);                                                        \
                                                                                                             \
  // Scale PoT down to 256 16 bit ints.                                                                      \
  //for (i=0; i<swi.analysis_cfg.gauss_pot_length; i++) gi->pot[i] = fp_PoT[i];   // ???                     \
  scale_factor = static_cast<float>(gi->g.max_power) / 255.0f;                                               \
  gi->g.pot.set_size(swi.analysis_cfg.gauss_pot_length);                                                     \
  float_to_uchar(fp_PoT, (unsigned char *)(gi->g.pot), swi.analysis_cfg.gauss_pot_length, scale_factor);

  if (!swi.analysis_cfg.gauss_null_chi_sq_thresh)
      swi.analysis_cfg.gauss_null_chi_sq_thresh=1.890;
  // Gauss score used for "best of" and graphics.
  // This score is now set to be based upon the probability that a signal
  // would occur due to noise and the probability that it is shaped like 
  // a Gaussian (normalized to 0 at thresholds).  Thanks to Tetsuji for
  // making me think about this. The Gaussian has 62 degrees of freedom and
  // the null hypothesis has 63 degrees of freedom when gauss_pot_length=64;
  //
  // Postpone the calculation unless we're sure we need the score.
  gauss_bins=swi.analysis_cfg.gauss_pot_length;
  gauss_dof=gauss_bins-2;
  null_dof=gauss_bins-1;
  
  analysis_state.FLOP_counter+=24.0;
#ifdef BOINC_APP_GRAPHICS
  if (chisq || graphics) {
#else
  if (chisq) {
#endif
    BOOLEAN newbest, report;
    score = lcgf(0.5*gauss_dof,std::max(ChiSq*0.5*gauss_bins,0.5*gauss_dof+1))
           -lcgf_gauss
           -lcgf(0.5*null_dof,std::max(null_ChiSq*0.5*gauss_bins,0.5*null_dof+1))
           +lcgf_null;
    newbest = chisq && (score > best_gauss->score);
    report = chisq && (PeakPower >= TrueMean * PoTInfo.GaussPeakPowerThresh) && (null_ChiSq >= swi.analysis_cfg.gauss_null_chi_sq_thresh);
#ifdef BOINC_APP_GRAPHICS
    if (newbest || report || graphics) {
#else
    if (newbest || report) {
#endif
      GAUSS_INFO gi;           
      gi.bin                  = bin;
      gi.fft_ind              = ifft;
      gi.g.chirp_rate 	  = ChirpFftPairs[analysis_state.icfft].ChirpRate;
      gi.g.fft_len    	  = ChirpFftPairs[analysis_state.icfft].FftLen;
      gi.g.sigma      	  = sigma;
      gi.g.peak_power 	  = PeakPower;
      gi.g.mean_power 	  = TrueMean;
      gi.g.chisqr     	  = ChiSq;
      gi.g.null_chisqr     	  = null_ChiSq;
      gi.g.freq       	  = cnvt_bin_hz(bin,  gi.g.fft_len);
      double t_offset=(((double)gi.fft_ind+0.5)/swi.analysis_cfg.gauss_pot_length)*
        PoTInfo.WUDuration;
      gi.g.detection_freq=calc_detection_freq(gi.g.freq,gi.g.chirp_rate,t_offset);
      gi.g.time       	  = swi.time_recorded+t_offset/86400.0;
      gi.g.max_power  = PoTMaxPower;
      gi.score        = score;
      time_to_ra_dec(gi.g.time, &gi.g.ra, &gi.g.decl);
      
      // Scale PoT down to 256 16 bit ints.
      scale_factor = static_cast<float>(gi.g.max_power) / 255.0f;
      gi.g.pot.set_size(swi.analysis_cfg.gauss_pot_length);
      float_to_uchar(fp_PoT, (unsigned char *)(gi.g.pot), swi.analysis_cfg.gauss_pot_length, scale_factor);
      
#ifdef BOINC_APP_GRAPHICS
      if (graphics)
        sah_graphics->gi.copy(&gi);
#endif
      if (newbest)
        *best_gauss = gi;
      if (report)
        return result_gaussian(gi);
    }
  }
  
//  // Only include "real" Gaussians (those meeting the chisqr threshold)
//  // in the best Gaussian display. 
//  if (ChiSq < swi.analysis_cfg.gauss_chi_sq_thresh) {
//    score = lcgf(0.5*gauss_dof,std::max(ChiSq*0.5*gauss_bins,0.5*gauss_dof+1))
//           -lcgf_gauss
//           -lcgf(0.5*null_dof,std::max(null_ChiSq*0.5*gauss_bins,0.5*null_dof+1))
//           +lcgf_null;
//    lazy_score = false;
//    if (score > best_gauss->score) {
//      INIT_GAUSSIAN;
//      lazy_gauss = false;
//      *best_gauss = *gi;
//    }
//  }
//  // Update gdata gauss info regardless of whether it is the
//  // best thus far or even passes the final threshold.  If
//  // a gaussian has made it this far, display it.
//#ifdef BOINC_APP_GRAPHICS
//  if (!nographics()) {
//    if (lazy_score) {
//      score = lcgf(0.5*gauss_dof,std::max(ChiSq*0.5*gauss_bins,0.5*gauss_dof+1))
//                -lcgf_gauss
//                -lcgf(0.5*null_dof,std::max(null_ChiSq*0.5*gauss_bins,0.5*null_dof+1))
//                +lcgf_null;
//      lazy_score = false;
//    }
//    if (lazy_gauss) {
//      INIT_GAUSSIAN;
//      lazy_gauss = false;
//    }
//    sah_graphics->gi.copy(gi);    
//  }
//#endif
//
//  analysis_state.FLOP_counter+=24.0;
//  // Final thresholding and reporting.
//  if (PeakPower / TrueMean >= PoTInfo.GaussPeakPowerThresh) {
//    if ((ChiSq <=  PoTInfo.GaussChiSqThresh) &&
//       (null_ChiSq >= swi.analysis_cfg.gauss_null_chi_sq_thresh))
//    {
//      if (lazy_score) {
//        score = lcgf(0.5*gauss_dof,std::max(ChiSq*0.5*gauss_bins,0.5*gauss_dof+1))
//                  -lcgf_gauss
//                  -lcgf(0.5*null_dof,std::max(null_ChiSq*0.5*gauss_bins,0.5*null_dof+1))
//                  +lcgf_null;
//        lazy_score = false;
//      }
//      if (lazy_gauss) {
//        INIT_GAUSSIAN;
//        lazy_gauss = false;
//      }
//      return result_gaussian(gi);
//    }
//  }
//  
//  if (!lazy_gauss) {
//    delete(gi);
//  }
  
  return 0;
}


int GaussFit(
  float * fp_PoT,
  int ul_FftLength,
  int ul_PoT
) {
  int i, retval;
  BOOLEAN b_IsAPeak;
  float f_NormMaxPower;
  float f_null_hyp;

  int ul_TOffset;
  int iSigma = static_cast<int>(floor(PoTInfo.GaussSigma+0.5));
  double diSigma = iSigma;
  float f_GroupSum, f_GroupMax;
  int i_f, iPeakLoc;

  float f_TotalPower,
        f_TrueMean,
        f_ChiSq,
        f_PeakPower,
        f_MaxPower;
  float f_MeanPower;
  float f_rMeanPower;
  float f_GaussSigma = PoTInfo.GaussSigma;
#ifdef USE_VECTORIZED_GAUSSIAN
  float f_TotalPowerOverTrueMeanElems;
#endif
#if (defined(__VEC__) || defined(__SSE3__)) && defined(USE_VECTORIZED_GAUSSIAN)
  vFloat v_TotalPower;
  vFloat v_rMeanPower;
  vFloat v_PowerThresh;
#ifdef __VEC__
  vFloat v_IsAPeak = ZERO;
  vFloat v_NormMaxPower = ZERO;
#else
  vFloat v_IsAPeak = _mm_setzero_ps();
  vFloat v_NormMaxPower = _mm_setzero_ps();
#endif
#endif
  
  // For setiathome the Sigma and Gaussian PoT length don't change during
  // a run of the application, so these frequently used values can be
  // precalculated and kept.
  static float f_PeakScaleFactor;
  static float *f_weight;
//#ifdef USE_VECTORIZED_GAUSSIAN
  static float *f_vweight;
  static float *f_PeakFilter;
  static float *f_BoxFilter;
  static float *f_corr;
  static float *f_swsum;
  static float f_rTrueMeanElems = 1.0f / (swi.analysis_cfg.gauss_pot_length - (4*iSigma+1));
//#endif

  if (!f_weight) {
    f_PeakScaleFactor = f_GetPeakScaleFactor(static_cast<float>(PoTInfo.GaussSigma));

//Jason: Shifying this downwards...
//#ifdef USE_VECTORIZED_GAUSSIAN
    // Altivec and SSE code uses double-sided f_weight to make memory accesses simpler
//    f_vweight = reinterpret_cast<float *>(malloc((2*PoTInfo.GaussTOffsetStop-1)*sizeof(float)));
    f_vweight = reinterpret_cast<float *>( malloc_a((2*PoTInfo.GaussTOffsetStop-1)*sizeof(float),MEM_ALIGN) );
    if (!f_vweight) SETIERROR(MALLOC_FAILED, "!f_vweight");
    f_vweight[PoTInfo.GaussTOffsetStop-1] = 1;
    
//    f_PeakFilter = reinterpret_cast<float *>(malloc((2*iSigma+1)*sizeof(float)));
    f_PeakFilter = reinterpret_cast<float *>( malloc_a((2*iSigma+1)*sizeof(float),MEM_ALIGN) );
    if (!f_PeakFilter) SETIERROR(MALLOC_FAILED, "!f_PeakFilter");
//    f_BoxFilter = reinterpret_cast<float *>(malloc((4*iSigma+1)*sizeof(float)));
    f_BoxFilter = reinterpret_cast<float *>( malloc_a((4*iSigma+1)*sizeof(float),MEM_ALIGN) );
    if (!f_BoxFilter) SETIERROR(MALLOC_FAILED, "!f_BoxFilter");
    f_PeakFilter[iSigma] = f_PeakScaleFactor;
    f_BoxFilter[2*iSigma] = f_rTrueMeanElems;
    
//    f_weight = reinterpret_cast<float *>(malloc(PoTInfo.GaussTOffsetStop*sizeof(float)));
    f_weight = reinterpret_cast<float *>( malloc_a(PoTInfo.GaussTOffsetStop*sizeof(float),MEM_ALIGN) );
    if (!f_weight) SETIERROR(MALLOC_FAILED, "!f_weight");
    f_weight[0] = 1;
    
//    f_corr = reinterpret_cast<float *>(malloc((PoTInfo.GaussTOffsetStop - PoTInfo.GaussTOffsetStart)*sizeof(float)));
//Works:    f_corr = reinterpret_cast<float *>( malloc_a((PoTInfo.GaussTOffsetStop - PoTInfo.GaussTOffsetStart)*sizeof(float),MEM_ALIGN));
      f_corr = reinterpret_cast<float *>( malloc_a((PoTInfo.GaussTOffsetStop - PoTInfo.GaussTOffsetStart + 4*iSigma)*sizeof(float),MEM_ALIGN));
//    f_swsum = reinterpret_cast<float *>(malloc((swi.analysis_cfg.gauss_pot_length-4*iSigma)*sizeof(float)));
//Works:    f_swsum = reinterpret_cast<float *>( malloc_a((swi.analysis_cfg.gauss_pot_length-4*iSigma)*sizeof(float),MEM_ALIGN));
    f_swsum = reinterpret_cast<float *>( malloc_a((swi.analysis_cfg.gauss_pot_length+4*iSigma)*sizeof(float),MEM_ALIGN));

//Jason: Shifying this downwards...
#ifdef USE_VECTORIZED_GAUSSIAN
    for (i = 1; i <= iSigma; i++) {
      float weight = static_cast<float>(EXP(i, 0, PoTInfo.GaussSigmaSq));
      f_vweight[PoTInfo.GaussTOffsetStop-1+i] = weight;
      f_vweight[PoTInfo.GaussTOffsetStop-1-i] = weight;
      f_PeakFilter[iSigma+i] = weight * f_PeakScaleFactor;
      f_PeakFilter[iSigma-i] = weight * f_PeakScaleFactor;
      f_BoxFilter[2*iSigma+i] = f_rTrueMeanElems;
      f_BoxFilter[2*iSigma-i] = f_rTrueMeanElems;
      
      f_weight[i] = weight;
    }
    
    for (; i <= 2*iSigma; i++) {
      float weight = static_cast<float>(EXP(i, 0, PoTInfo.GaussSigmaSq));
      f_vweight[PoTInfo.GaussTOffsetStop-1+i] = weight;
      f_vweight[PoTInfo.GaussTOffsetStop-1-i] = weight;
      f_BoxFilter[2*iSigma+i] = f_rTrueMeanElems;
      f_BoxFilter[2*iSigma-i] = f_rTrueMeanElems;
      
      f_weight[i] = weight;      
    }

    for (; i < PoTInfo.GaussTOffsetStop; i++) {
      float weight = static_cast<float>(EXP(i, 0, PoTInfo.GaussSigmaSq));
      f_vweight[PoTInfo.GaussTOffsetStop-1+i] = weight;
      f_vweight[PoTInfo.GaussTOffsetStop-1-i] = weight;

      f_weight[i] = weight;
    }    
#else
//    f_weight = reinterpret_cast<float *>(malloc(PoTInfo.GaussTOffsetStop*sizeof(float)));
    f_weight = reinterpret_cast<float *>(malloc_a(PoTInfo.GaussTOffsetStop*sizeof(float), MEM_ALIGN) );
    if (!f_weight) SETIERROR(MALLOC_FAILED, "!f_weight");
    for (i = 0; i < PoTInfo.GaussTOffsetStop; i++) {
      f_weight[i] = static_cast<float>(EXP(i, 0, PoTInfo.GaussSigmaSq));
    }
#endif
//Jason Added
//    for (i = 0; i < PoTInfo.GaussTOffsetStop; i++) {
//      f_weight[i] = static_cast<float>(EXP(i, 0, PoTInfo.GaussSigmaSq));
//    }
//Jason Ends
  }
  
#if defined(__VEC__) && defined(USE_VECTORIZED_GAUSSIAN)
  // Find reciprocal of mean over power-of-time array
  // Assumes that swi.analysis_cfg.gauss_pot_length is a multiple of 4
  v_TotalPower = ZERO;
  for (i = 0; i < swi.analysis_cfg.gauss_pot_length; i += 4) {
    const float *p = fp_PoT + i;
    vector float v_PoT = vec_ld(0, p);
    v_TotalPower = vec_add(v_TotalPower, v_PoT);
  }
  v_TotalPower = vec_add(v_TotalPower, vec_sld(v_TotalPower, v_TotalPower, 8));
  v_TotalPower = vec_add(v_TotalPower, vec_sld(v_TotalPower, v_TotalPower, 4));
  v_rMeanPower = vec_madd(vec_recip(v_TotalPower),
                         vec_fpsplat((float) swi.analysis_cfg.gauss_pot_length), ZERO);
  
  // Normalize power-of-time and check for the existence
  // of at least 1 peak.
  // grab the maximum, while we're at it
  b_IsAPeak = false;
  v_PowerThresh = vec_fpsplat(PoTInfo.GaussPowerThresh);
  for (i = 0; i < swi.analysis_cfg.gauss_pot_length; i += 4) {
    float *p = fp_PoT + i;
    vector float v_PoT = vec_ld(0, p);
    v_PoT = vec_madd(v_PoT, v_rMeanPower, ZERO);
    v_IsAPeak = vec_or(v_IsAPeak, vec_cmpgt(v_PoT, v_PowerThresh));
	v_NormMaxPower = vec_max(v_NormMaxPower, v_PoT);
    vec_st(v_PoT, 0, p);
  }
  b_IsAPeak = vec_any_ne(v_IsAPeak, ZERO);
  v_NormMaxPower = vec_max(v_NormMaxPower, vec_sld(v_NormMaxPower, v_NormMaxPower, 8));
  v_NormMaxPower = vec_max(v_NormMaxPower, vec_sld(v_NormMaxPower, v_NormMaxPower, 4));
#elif defined(__SSE3__) && defined(USE_VECTORIZED_GAUSSIAN)
  // Find reciprocal of mean over power-of-time array
  // Assumes that swi.analysis_cfg.gauss_pot_length is a multiple of 4
  v_TotalPower = _mm_setzero_ps();
  for (i = 0; i < swi.analysis_cfg.gauss_pot_length; i += 4) {
    v_TotalPower = _mm_add_ps(v_TotalPower, _mm_load_ps(fp_PoT+i));
  }
  v_TotalPower = _mm_hadd_ps(v_TotalPower, v_TotalPower);
  v_TotalPower = _mm_hadd_ps(v_TotalPower, v_TotalPower);
//  // don't use movehl here, since the whole vector needs to be right
//  v_TotalPower = _mm_add_ps(v_TotalPower, _mm_movehl_ps(v_TotalPower, v_TotalPower));
//  v_TotalPower = _mm_hadd_ps(v_TotalPower, v_TotalPower);
  v_rMeanPower = _mm_mul_ps(vec_recip(v_TotalPower), _mm_set1_ps((float) swi.analysis_cfg.gauss_pot_length));
  
  // Normalize power-of-time and check for the existence
  // of at least 1 peak.
  v_PowerThresh = _mm_set1_ps(PoTInfo.GaussPowerThresh);
  for (i = 0; i < swi.analysis_cfg.gauss_pot_length; i += 4) {
    float *p = fp_PoT + i;
    vFloat v_PoT = _mm_load_ps(p);
    v_PoT = _mm_mul_ps(v_PoT, v_rMeanPower);
    v_IsAPeak = _mm_or_ps(v_IsAPeak, _mm_cmpgt_ps(v_PoT, v_PowerThresh));
    v_NormMaxPower = _mm_max_ps(v_NormMaxPower, v_PoT);
    _mm_store_ps(p, v_PoT);
  }  
  // assign b_IsAPeak here
  b_IsAPeak = _mm_movemask_ps(v_IsAPeak);
  v_NormMaxPower = _mm_max_ps(v_NormMaxPower, _mm_movehl_ps(v_NormMaxPower, v_NormMaxPower));
  v_NormMaxPower = _mm_max_ps(v_NormMaxPower, _mm_movehdup_ps(v_NormMaxPower));
  
#else
  // Find mean over power-of-time array
  f_TotalPower = 0;
  for (i = 0; i < swi.analysis_cfg.gauss_pot_length; i++) {
    f_TotalPower += fp_PoT[i];
  }
  f_MeanPower = f_TotalPower / swi.analysis_cfg.gauss_pot_length;
  f_rMeanPower = 1.0f / f_MeanPower;

  // Normalize power-of-time and check for the existence
  // of at least 1 peak.
  b_IsAPeak = false;
  for (i = 0; i < swi.analysis_cfg.gauss_pot_length; i++) {
    fp_PoT[i] *= f_rMeanPower;
    if (fp_PoT[i] > PoTInfo.GaussPowerThresh) b_IsAPeak = true;
  }
#endif
  analysis_state.FLOP_counter+=3.0*swi.analysis_cfg.gauss_pot_length+2;

  if (!b_IsAPeak) {
    //printf("no peak\n");
    return 0;  // no peak - bail on this PoT
  }

  // Recalculate Total Power across normalized array.
  // Given that powers are positive, f_TotalPower will
  // end up being == swi.analysis_cfg.gauss_pot_length.
  //f_TotalPower   = 0;
  //f_NormMaxPower = 0;
  //// Also locate group with the highest sum for use in second bailout check.
  //f_GroupSum     = 0;
  //f_GroupMax     = 0;
  //for (i = 0, i_f = -iSigma; i < swi.analysis_cfg.gauss_pot_length; i++, i_f++) {
  //  f_TotalPower += fp_PoT[i];
  //  if(fp_PoT[i] > f_NormMaxPower) f_NormMaxPower = fp_PoT[i];
  //  f_GroupSum += fp_PoT[i] - ((i_f < 0) ? 0.0f : fp_PoT[i_f]);
  //  if (f_GroupSum > f_GroupMax) {
  //    f_GroupMax = f_GroupSum;
  //    iPeakLoc = i - iSigma/2;
  //  }
  //}
  
  // let's not and say we did
  f_TotalPower = swi.analysis_cfg.gauss_pot_length;
  f_GroupSum = 0;
#if (defined(__VEC__) || defined(__SSE3__)) && defined(USE_VECTORIZED_GAUSSIAN)
  {
	float *vp_NormMaxPower = (float *) &v_NormMaxPower;
	f_NormMaxPower = *vp_NormMaxPower;
  }
#endif
  
  // running sums don't really vectorize
  // initialize sum and maximum
#ifndef USE_VECTORIZED_GAUSSIAN
  f_NormMaxPower = 0;
#endif
  for (i = 0; i < iSigma; i++) {
	f_GroupSum += fp_PoT[i];
  }
  f_GroupMax = f_GroupSum;
  iPeakLoc = i - iSigma/2;
  for ( ; i < swi.analysis_cfg.gauss_pot_length; i++) {
	f_GroupSum += fp_PoT[i] - fp_PoT[i - iSigma];
#ifndef USE_VECTORIZED_GAUSSIAN
	if (fp_PoT[i] > f_NormMaxPower) f_NormMaxPower = fp_PoT[i];
#endif
	if (f_GroupSum > f_GroupMax) {
	  f_GroupMax = f_GroupSum;
	  iPeakLoc = i - iSigma/2;
	}
  }
  
#ifdef USE_VECTORIZED_GAUSSIAN
  f_TotalPowerOverTrueMeanElems = f_TotalPower * f_rTrueMeanElems;
#endif
  
  // Check at the group peak location whether data may contain Gaussians
  // (but only after the first hurry-up Gaussian has been set for graphics)

  if (best_gauss->display_power_thresh != 0) {
    iPeakLoc = std::max(PoTInfo.GaussTOffsetStart,
                       (std::min(PoTInfo.GaussTOffsetStop - 1, iPeakLoc)));
  
#if (defined(__VEC__) || defined(__SSE3__)) && defined(USE_VECTORIZED_GAUSSIAN)
    f_TrueMean = v_GetTrueMean(
                   fp_PoT,
                   swi.analysis_cfg.gauss_pot_length,
                   f_TotalPower,
                   iPeakLoc,
                   2 * iSigma
                 );
#else
    f_TrueMean = f_GetTrueMean(
                   fp_PoT,
                   swi.analysis_cfg.gauss_pot_length,
                   f_TotalPower,
                   iPeakLoc,
                   2 * iSigma
                 );
#endif
  
    f_PeakPower = f_GetPeak(
                    fp_PoT,
                    iPeakLoc,
                    iSigma,
                    f_TrueMean,
                    f_PeakScaleFactor,
                    f_vweight
                  );

    analysis_state.FLOP_counter+=5.0*swi.analysis_cfg.gauss_pot_length+5;

    if (f_PeakPower < f_TrueMean * best_gauss->display_power_thresh*0.5f) {
      return 0;  // not even a weak peak at max group - bail on this PoT
    }
  }

#ifdef USE_VECTORIZED_GAUSSIAN
//  conv(fp_PoT, 1, f_PeakFilter, 1, f_corr, 1,
//       PoTInfo.GaussTOffsetStop - PoTInfo.GaussTOffsetStart, 2*iSigma+1);
//  ippsCrossCorr_32f(fp_PoT, PoTInfo.GaussTOffsetStop - PoTInfo.GaussTOffsetStart,
//	             f_PeakFilter, 2*iSigma+1, f_corr, PoTInfo.GaussTOffsetStop - PoTInfo.GaussTOffsetStart, 0);
  ippsConv_32f(fp_PoT, PoTInfo.GaussTOffsetStop - PoTInfo.GaussTOffsetStart+2*iSigma,  f_PeakFilter, 2*iSigma+1, f_corr);
//  conv(fp_PoT, 1, f_BoxFilter, 1, f_swsum, 1,
//       swi.analysis_cfg.gauss_pot_length-4*iSigma, 4*iSigma+1);
//  ippsCrossCorr_32f(fp_PoT, swi.analysis_cfg.gauss_pot_length-4*iSigma,
//	             f_BoxFilter, 4*iSigma+1, f_swsum, swi.analysis_cfg.gauss_pot_length-4*iSigma, 0);
  ippsConv_32f(fp_PoT, swi.analysis_cfg.gauss_pot_length,
	             f_BoxFilter, 4*iSigma+1, f_swsum);

#endif

  // slide dynamic gaussian across the Power Of Time array
  float display_power_thresh = best_gauss->display_power_thresh;
#ifdef USE_VECTORIZED_GAUSSIAN
  float f_tempTotalPowerOverTrueMeanElems = f_TotalPowerOverTrueMeanElems;
//  float *f_tcorr = f_corr;    // copy static array pointers
  float *f_tcorr = f_corr+2*iSigma;    // copy static array pointers
//  float *f_tswsum = f_swsum;
  float *f_tswsum = f_swsum+4*iSigma;

#endif
  for (ul_TOffset = PoTInfo.GaussTOffsetStart;
       ul_TOffset < PoTInfo.GaussTOffsetStop;
       ul_TOffset++
      ) {

    // TrueMean is the mean power of the data set minus all power
    // out to 2 sigma from our current TOffset.
#if (defined(__VEC__) || defined(__SSE3__)) && defined(USE_VECTORIZED_GAUSSIAN)

//Jason: Reenabled if for correelation
    if ((ul_TOffset - 2 * iSigma) < 0 || (ul_TOffset + 2 * iSigma) >= swi.analysis_cfg.gauss_pot_length) {
      f_TrueMean = v_GetTrueMean(
                     fp_PoT,
                     swi.analysis_cfg.gauss_pot_length,
                     f_TotalPower,
                     ul_TOffset,
                     2 * iSigma
                   );
   } else { //reenabled else case for correlation
      f_TrueMean = f_tempTotalPowerOverTrueMeanElems - f_tswsum[ul_TOffset - 2 * iSigma];
//  f_TrueMean = f_GetTrueMean(
//  f_TrueMean = v_GetTrueMean(
//                   fp_PoT,
//                   swi.analysis_cfg.gauss_pot_length,
//                   f_TotalPower,
//                   ul_TOffset,
//                   2 * iSigma
//                 );
}
#else
  f_TrueMean = f_GetTrueMean(
                   fp_PoT,
                   swi.analysis_cfg.gauss_pot_length,
                   f_TotalPower,
                   ul_TOffset,
                   2 * iSigma
                 );
#endif

#ifdef USE_VECTORIZED_GAUSSIAN
 f_PeakPower = f_tcorr[ul_TOffset-iSigma] - f_TrueMean;
//	fprintf(stderr,"PPxC: %8f " , f_PeakPower);
#else
// Jason: Trial Replacement
  float f_PeakPower = f_GetPeak(
                    fp_PoT,
                    ul_TOffset,
                    iSigma,
                    f_TrueMean,
                    f_PeakScaleFactor,
					f_vweight
                  );

#endif

    // worth looking at ?
    if (f_PeakPower < f_TrueMean * display_power_thresh) {
      continue;
    }

    // bump up the display threshold to its final value.
    // We could bump it up only to the gaussian just found,
    // but that would cause a lot of time waste
    // computing chisq etc.
    if (display_power_thresh == 0) {
    	display_power_thresh = best_gauss->display_power_thresh = PoTInfo.GaussPeakPowerThresh/3;
    }

#ifdef TEXT_UI
    if(dump_pot) {
      printf("Found good peak at this PoT element.... truemean is %f, power is %f\n", f_TrueMean, f_PeakPower);
    }
#endif

#ifdef DUMP_GAUSSIAN
    gul_PoT = ul_PoT;
    gul_Fftl = ul_FftLength;
#endif

    // look at it - try to fit
#if (defined(__VEC__) || defined(__SSE3__)) && defined(USE_VECTORIZED_GAUSSIAN)

	f_ChiSq = v_GetChiSq(
                fp_PoT,
                swi.analysis_cfg.gauss_pot_length,
                ul_TOffset,
                f_PeakPower,
                f_TrueMean,
                f_vweight,
                &f_null_hyp
              );
#else
    f_ChiSq = f_GetChiSq(
                fp_PoT,
                swi.analysis_cfg.gauss_pot_length,
                ul_TOffset,
                f_PeakPower,
                f_TrueMean,
                f_weight,
                &f_null_hyp
              );
#endif
    
#ifdef TEXT_UI
    if(dump_pot) {
      printf("Checking ChiSqr for PoT dump....\n");
      if(f_ChiSq <= PoTInfo.GaussChiSqThresh) {
        int dump_i;
        printf(
          "POT %d %f %f %f %f %d ",
          ul_TOffset,
          f_PeakPower,
          f_TrueMean,
          PoTInfo.GaussSigmaSq,
          f_ChiSq,
          ul_PoT
        );
        for (dump_i = 0; dump_i < swi.analysis_cfg.gauss_pot_length; dump_i++) {
          printf("%f ", fp_PoT[dump_i]);
        }
        printf("\n");
      } else {
        printf("ChiSqr is %f, not good enough.\n", f_ChiSq);
      }
    }
#endif

    retval = ChooseGaussEvent(
               ul_TOffset,
               f_PeakPower,
               f_TrueMean,
               f_ChiSq,
	       f_null_hyp,
               ul_PoT,
               f_GaussSigma,
               f_NormMaxPower,
               fp_PoT
             );
    if (retval) SETIERROR(retval,"from ChooseGaussEvent");

  } // End of sliding gaussian

#ifdef USE_VECTORIZED_GAUSSIAN
//#if 0
  int ftm_start = (PoTInfo.GaussTOffsetStart > 2*iSigma) ? PoTInfo.GaussTOffsetStart : 2*iSigma;
  int ftm_lim   = (PoTInfo.GaussTOffsetStop < swi.analysis_cfg.gauss_pot_length-2*iSigma) ?
    PoTInfo.GaussTOffsetStop : swi.analysis_cfg.gauss_pot_length-2*iSigma;
  analysis_state.FLOP_counter+=
    (PoTInfo.GaussTOffsetStop-PoTInfo.GaussTOffsetStart)*6.0*diSigma+
    (ftm_lim-ftm_start)*(4.0*diSigma+6.0);
#endif

  return 0;

} // End of gaussfit()


