// Copyright 2003 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

// In addition, as a special exception, the Regents of the University of
// California give permission to link the code of this program with libraries
// that provide specific optimized fast Fourier transform (FFT) functions and
// distribute a linked executable.  You must obey the GNU General Public 
// License in all respects for all of the code used other than the FFT library
// itself.  Any modification required to support these libraries must be
// distributed in source code form.  If you modify this file, you may extend 
// this exception to your version of the file, but you are not obligated to 
// do so. If you do not wish to do so, delete this exception statement from 
// your version.


#include "config.h"

#include <stdio.h>
#include <iostream>
#include <string.h>
#include <string>
#include <stdlib.h>
#include <math.h>

#ifdef BOINC_APP_GRAPHICS
#include "graphics_api.h"
#ifdef DYNAMIC_GRAPHICS
#include "graphics_lib.h"
#endif
#endif

#ifdef __SSE3__
#pragma message("---- SSE3 ----")
#include <pmmintrin.h>
#elif defined __SSE2__
#pragma message("---- SSE2 ----")
#include <emmintrin.h>
#elif defined __SSE__
#pragma message("---- SSE ----")
#include <xmmintrin.h>
#endif

#ifdef USE_SSE2
#pragma message("------> Enabling SSE3 emulation by SSE2 ----")
#include "SSE3_emulate.h"
#endif


#include "util.h"
#include "s_util.h"
#include "analyze.h"
#include "gaussfit.h"
#include "seti.h"
#include "chirpfft.h"
#include "analyzeReport.h"
#include "analyzePoT.h"

#include "analyzefuncs.h"

#ifdef BOINC_APP_GRAPHICS
#include "sah_gfx.h"
#endif

#include "../db/schema_master.h"

//#include <Accelerate/Accelerate.h>

#if defined(__VEC__) || defined(__SSE3__)
//#define ZERO		((vFloat) {0.0f, 0.0f, 0.0f, 0.0f})
//#define IZERO		((vUInt32) {0, 0, 0, 0})
//#define IFOUR		((vUInt32) {4, 4, 4, 4})
   #define ZERO		(_mm_set_ps(0.0f, 0.0f, 0.0f, 0.0f))
   #define IZERO	(_mm_set_epi32(0, 0, 0, 0))
   #define IFOUR	(_mm_set_epi32(4, 4, 4, 4))
#endif

#ifdef USE_PPC_OPTIMIZATIONS
#define USE_VECTORIZED_SPIKE
#elif defined(USE_I386_OPTIMIZATIONS)
#define USE_VECTORIZED_SPIKE
#endif

#ifdef __VEC__
inline vector float vec_fpsplat(float f_Constant) {
  vector unsigned char vuc_Splat;
  vector float vf_Constant;
  
  vuc_Splat = vec_lvsl(0, &f_Constant);
  vf_Constant = vec_lde(0, &f_Constant);
  vuc_Splat = (vector unsigned char) vec_splat((vector float) vuc_Splat, 0);
  vf_Constant = vec_perm(vf_Constant, vf_Constant, vuc_Splat);
  
  return vf_Constant;
}
#endif

int FindSpikes(
  float * fp_PowerSpectrum,
  int ul_NumDataPoints,
  int fft_num,
  SETI_WU_INFO& swi,
  float * fp_MaxNoDC,
  unsigned int * ulp_MaxInd,
  float * fp_Mean
) {
  //int i, j, k, m, bin, retval;
  int i, j, k, m, retval;
  float temp;
  //float total, MeanPower, spike_score;
  float total, MeanPower;
  SPIKE_INFO si;
#if defined(__VEC__) && defined(USE_VECTORIZED_SPIKE)
  vFloat vtemptotal;
#endif
#if (defined(__VEC__) || defined(__SSE3__)) && defined(USE_VECTORIZED_SPIKE)
  vFloat vtotal;
#endif

  i = j = k = m = 0;
  total = 0.0;

  if (!fp_Mean) {
#if defined(__VEC__) && defined(USE_VECTORIZED_SPIKE)
	vtemptotal = ZERO;
	for(i = 0; i < ul_NumDataPoints; i += 4) {
	  const float *p = (const float *) (fp_PowerSpectrum + i);
	  vector float p1 = vec_ld(0, p);
	  vtemptotal = vec_add(vtemptotal, p1);
	}
	vtotal = vtemptotal;
	float *vptotal = (float *) &vtotal;
	total = (vptotal[0] + vptotal[1]) + (vptotal[2] + vptotal[3]);
	MeanPower = total / ul_NumDataPoints;
#elif defined(__SSE3__) && defined(USE_VECTORIZED_SPIKE)
	vtotal = _mm_setzero_ps();
	for(i = 0; i < ul_NumDataPoints; i += 4) {
	  vtotal = _mm_add_ps(vtotal, _mm_load_ps(fp_PowerSpectrum+i));
	}
  //  vtotal = _mm_hadd_ps(vtotal, vtotal);
  //  vtotal = _mm_hadd_ps(vtotal, vtotal);
	vtotal = _mm_add_ps(vtotal, _mm_movehl_ps(vtotal, vtotal));
	vtotal = _mm_add_ss(vtotal, _mm_movehdup_ps(vtotal));
//	_mm_store_ss(&total, vtotal);
  total = _mm_cvtss_f32(vtotal);
	MeanPower = total / ul_NumDataPoints;
#elif defined(USE_VECTORIZED_SPIKE)
	vDSP_meanv(fp_PowerSpectrum, 1, &MeanPower, ul_NumDataPoints);
#else
	for(i = 0; i < ul_NumDataPoints; i++) {
	  total += fp_PowerSpectrum[i];
	}
	MeanPower = total / ul_NumDataPoints;
#endif
  } else MeanPower = *fp_Mean;
  
  // Here we extract the spikes_to_report highest power events,
  // outputing them as we go.
  // Index usage:
  // i : walk power spectrum us_NumToReport times
  // j : walks power spectrum for each i
  // k : marks current high power candidate while j walks on
  // m : marks the current tail of the high power hit "list"

  for (i = 0; i < swi.analysis_cfg.spikes_per_spectrum; i++) {

    temp = 0.0;

    // Walk the array, looking for the first/next highest power.
    // Start j at 1, in order to skip the DC (ie 0) bin.
    // NOTE: this is a simple scan for high powers.  Nice and fast
    // for a very low i.  If i is substantial, this code should be
    // replaced with an index (q)sort.  Do *not* sort the power
    // spectrum itself in place - it's used elsewhere.
#if defined(__VEC__) && defined(USE_VECTORIZED_SPIKE)
    // this imposes a rather large fixed overhead
    // fall back to scalar for small arrays?
    vector float vpower, vtempmax, vmax, vmmax;
    vector unsigned int vj, vk, vtempk, vcmp;
    
    if (m != 0) {
      vmmax = vec_fpsplat(fp_PowerSpectrum[m]);
      vtempk = ((vUInt32) {0, 1, 2, 3});
      vtempmax = vec_ld(0, fp_PowerSpectrum);
      vtempmax = vec_and(vtempmax, (vector float) {0, -1, -1, -1});
      vcmp = (vector unsigned int) vec_cmplt(vtempmax, vmmax);
      vtempmax = vec_and(vtempmax, (vFloat) vcmp);
      vtempk = vec_and(vtempk, vcmp);
      
      vj = ((vector unsigned int) {4, 5, 6, 7});
      for (j = 4; j < ul_NumDataPoints; j += 4) {
        const float *p = fp_PowerSpectrum + j;
        vpower = vec_ld(0, p);
        vcmp = vec_and((vector unsigned int) vec_cmplt(vpower, vmmax),
                       (vector unsigned int) vec_cmpgt(vpower, vtempmax));
        vtempmax = vec_sel(vtempmax, vpower, vcmp);
        vtempk = vec_sel(vtempk, vj, vcmp);
        vj = vec_add(vj, IFOUR);
      }
	} else if (!fp_MaxNoDC) {
      // initialize k / handle elements 1-3
      vtempk = ((vector unsigned int) {0, 1, 2, 3});
      vtempmax = vec_ld(0, fp_PowerSpectrum);
      vtempmax = vec_and(vtempmax, (vFloat) ((vector unsigned int) {0, 0xffffffff, 0xffffffff, 0xffffffff}));
      
      vj = ((vector unsigned int) {4, 5, 6, 7});
      for (j = 4; j < ul_NumDataPoints; j += 4) {
        const float *p = (const float *) fp_PowerSpectrum + j;
        vpower = vec_ld(0, p);
        vcmp = (vector unsigned int) vec_cmpgt(vpower, vtempmax);
        vtempmax = vec_max(vpower, vtempmax);
        vtempk = vec_sel(vtempk, vj, vcmp);
        vj = vec_add(vj, IFOUR);
      }
    }
    
    // get the index corresponding to the maximum of the vector
	// handle tiebreakers by picking the lower index, since that's what the scalar version does
//    vcmp = (vector unsigned int) vec_cmplt(vtempmax, vec_sld(vtempmax, vtempmax, 8));
//    vtempmax = vec_max(vtempmax, vec_sld(vtempmax, vtempmax, 8));
//    vtempk = vec_sel(vtempk, vec_sld(vtempk, vtempk, 8), vcmp);
//    vcmp = (vector unsigned int) vec_cmplt(vtempmax, vec_sld(vtempmax, vtempmax, 4));
//    vmax = vec_max(vtempmax, vec_sld(vtempmax, vtempmax, 4));
//    vk = vec_sel(vtempk, vec_sld(vtempk, vtempk, 4), vcmp);
	if (!fp_MaxNoDC) {
	  vcmp = (vector unsigned int) vec_or(vec_cmpgt(vtempmax, vec_sld(vtempmax, vtempmax, 8)),
										  vec_and(vec_cmpeq(vtempmax, vec_sld(vtempmax, vtempmax, 8)),
												  vec_cmplt(vtempk, vec_sld(vtempk, vtempk, 8))));
	  vtempmax = vec_sel(vec_sld(vtempmax, vtempmax, 8), vtempmax, vcmp);
	  vtempk = vec_sel(vec_sld(vtempk, vtempk, 8), vtempk, vcmp);
	  vcmp = (vector unsigned int) vec_or(vec_cmpgt(vtempmax, vec_sld(vtempmax, vtempmax, 4)),
										  vec_and(vec_cmpeq(vtempmax, vec_sld(vtempmax, vtempmax, 4)),
												  vec_cmplt(vtempk, vec_sld(vtempk, vtempk, 4))));
	  vmax = vec_sel(vec_sld(vtempmax, vtempmax, 4), vtempmax, vcmp);
	  vk = vec_sel(vec_sld(vtempk, vtempk, 4), vtempk, vcmp);
	  float *vtempVp = (float *) &vmax;
	  int *vkVp = (int *) &vk;
	  temp = *vtempVp;
	  k = *vkVp;
	} else {
	  temp = *fp_MaxNoDC;
	  k = *ulp_MaxInd;
	}
#elif defined(__SSE2__) && defined(USE_VECTORIZED_SPIKE)
    float stemp = temp;
    int sk = k;
    
    // this imposes a rather large fixed overhead
    // fall back to scalar for small arrays or hook IPP's min function where possible?
    vFloat vpower, vmax, vcmp, vmmax;
    vUInt32 vj, vk;
    
    if (m != 0) {
      // initialize k / handle elements 1-3
      // this doesn't seem to happen in the general case
      vmmax = _mm_load_ss(fp_PowerSpectrum+m);
      vmmax = _mm_shuffle_ps(vmmax, vmmax, 0x0);
      vk = _mm_set_epi32(3, 2, 1, 0);
      vmax = _mm_load_ps(fp_PowerSpectrum);
      vmax = _mm_and_ps(vmax, (vFloat) _mm_set_epi32(-1, -1, -1, 0));
      vcmp = _mm_cmplt_ps(vmax, vmmax);
      vmax = _mm_and_ps(vmax, vcmp);
      vk = _mm_and_si128(vk, (vUInt32) vcmp);
      
      vj = _mm_set_epi32(7, 6, 5, 4);
      for (j = 4; j < ul_NumDataPoints; j += 4) {
        vpower = _mm_load_ps(fp_PowerSpectrum+j);
        vcmp = _mm_and_ps(_mm_cmplt_ps(vpower, vmmax), _mm_cmpgt_ps(vpower, vmax));
        vmax = _mm_or_ps(_mm_and_ps(vpower, vcmp),
                         _mm_andnot_ps(vcmp, vmax));
        vk = _mm_or_si128(_mm_and_si128(vj, (vUInt32) vcmp),
                          _mm_andnot_si128((vUInt32) vcmp, vk));
        vj = _mm_add_epi32(vj, IFOUR);
      }
    } else if (!fp_MaxNoDC) {
      // initialize k / handle elements 1-3
      vk = _mm_set_epi32(3, 2, 1, 0);
      vmax = _mm_load_ps(fp_PowerSpectrum);
      vmax = _mm_and_ps(vmax, (vFloat) _mm_set_epi32(-1, -1, -1, 0));
      
      vj = _mm_set_epi32(7, 6, 5, 4);
      for (j = 4; j < ul_NumDataPoints; j += 4) {
        vpower = _mm_load_ps(fp_PowerSpectrum+j);
        vcmp = _mm_cmpgt_ps(vpower, vmax);
        vmax = _mm_or_ps(_mm_and_ps(vpower, vcmp),
                         _mm_andnot_ps(vcmp, vmax));
        vk = _mm_or_si128(_mm_and_si128(vj, (vUInt32) vcmp),
                          _mm_andnot_si128((vUInt32) vcmp, vk));
        vj = _mm_add_epi32(vj, IFOUR);
      }      
    }
    
    if (!fp_MaxNoDC) {
    // get the index corresponding to the maximum of the vector
    // handle tiebreakers by picking the lower index, since that's what the scalar version does
//    vcmp = _mm_cmpgt_ps(vmax, _mm_shuffle_ps(vmax, vmax, 0x4e));
//    vmax = _mm_max_ps(vmax, _mm_shuffle_ps(vmax, vmax, 0x4e));
//    vk = _mm_or_si128(_mm_and_si128(vk, (vUInt32) vcmp),
//                      _mm_andnot_si128((vUInt32) vcmp, _mm_shuffle_epi32(vk, 0x4e)));
//    vcmp = _mm_cmpgt_ps(vmax, _mm_shuffle_ps(vmax, vmax, 0x39));
//    vmax = _mm_max_ps(vmax, _mm_shuffle_ps(vmax, vmax, 0x39));
//    vk = _mm_or_si128(_mm_and_si128(vk, (vUInt32) vcmp),
//                      _mm_andnot_si128((vUInt32) vcmp, _mm_shuffle_epi32(vk, 0x39)));
    vcmp = _mm_or_ps(_mm_cmpgt_ps(vmax, _mm_movehl_ps(vmax, vmax)),
                     _mm_and_ps(_mm_cmpeq_ps(vmax, _mm_movehl_ps(vmax, vmax)),
                                (vFloat) _mm_cmplt_epi32(vk, (vUInt32) _mm_movehl_ps((vFloat) vk, (vFloat) vk))));
    vmax = _mm_or_ps(_mm_and_ps(vmax, vcmp),
                   _mm_andnot_ps(vcmp, _mm_movehl_ps(vmax, vmax)));
    vk = _mm_or_si128(_mm_and_si128(vk, (vUInt32) vcmp),
                      _mm_andnot_si128((vUInt32) vcmp, (vUInt32) _mm_movehl_ps((vFloat) vk, (vFloat) vk)));
//    vcmp = _mm_cmpgt_ps(vmax, _mm_shuffle_ps(vmax, vmax, 0x1));
//    vmax = _mm_max_ps(vmax, _mm_shuffle_ps(vmax, vmax, 0x1));
//    vk = _mm_or_si128(_mm_and_si128(vk, (vUInt32) vcmp),
//                      _mm_andnot_si128((vUInt32) vcmp, _mm_shuffle_epi32(vk, 0x1)));
    vcmp = _mm_or_ps(_mm_cmpgt_ps(vmax, _mm_movehdup_ps(vmax)),
                     _mm_and_ps(_mm_cmpeq_ps(vmax, _mm_movehdup_ps(vmax)),
                                (vFloat) _mm_cmplt_epi32(vk, (vUInt32) _mm_movehdup_ps((vFloat) vk))));
    vmax = _mm_or_ps(_mm_and_ps(vmax, vcmp),
                     _mm_andnot_ps(vcmp, _mm_movehdup_ps(vmax)));
    vk = _mm_or_si128(_mm_and_si128(vk, (vUInt32) vcmp),
                      _mm_andnot_si128((vUInt32) vcmp, (vUInt32) _mm_movehdup_ps((vFloat) vk)));
//    _mm_store_ss(&temp, vmax);
    temp = _mm_cvtss_f32(vmax);
//    _mm_store_ss((float *) &k, (vFloat) vk); // XXX: is there a less hacky way to do this?
    k = _mm_cvtsi128_si32(vk);
  } else {
  	  temp = *fp_MaxNoDC;
  	  k = *ulp_MaxInd;
  }
#else
    for (j = 1; j < ul_NumDataPoints; j++) {

      if (fp_PowerSpectrum[j] > temp) {
        if (fp_PowerSpectrum[j] < fp_PowerSpectrum[m] || m == 0) {
          temp = fp_PowerSpectrum[j];
          k = j;
        }
      }
    } // temp now = first/next highest power and k = it's bin number
#endif

    m = k; 		// save the "lowest" highest.

    //  spike info
    si.s.peak_power 	 = temp/MeanPower;
    si.s.mean_power	 = 1.0;
    si.bin 		 = k;
    si.fft_ind 		 = fft_num;	
    si.s.chirp_rate 	 = ChirpFftPairs[analysis_state.icfft].ChirpRate;
    si.s.fft_len    	 = ChirpFftPairs[analysis_state.icfft].FftLen;
    si.s.freq		 = cnvt_bin_hz(si.bin,  si.s.fft_len);
    double t_offset=((double)si.fft_ind+0.5)*(double)si.s.fft_len/
          swi.subband_sample_rate;
    si.s.detection_freq=calc_detection_freq(si.s.freq,si.s.chirp_rate,t_offset);
    si.s.time		 = swi.time_recorded + t_offset / 86400.0;
    time_to_ra_dec(si.s.time, &si.s.ra, &si.s.decl);

    // Score used for "best of" and graphics.
    si.score 		 = si.s.peak_power / SPIKE_SCORE_HIGH;
    si.score 	  	 = si.score > 0.0f ? (float)log10(si.score) : 0.0f;
    // if best_spike.s.fft_len == 0, there is not yet a first spike
    if (si.score > best_spike->score || best_spike->s.fft_len == 0) {
      *best_spike 			= si;
#ifdef BOINC_APP_GRAPHICS
      if (!nographics()) sah_graphics->si.copy(&si);
#endif
    }

    // Report a signal if it excceeds threshold.
    if (si.s.peak_power > (swi.analysis_cfg.spike_thresh)) {
      retval = result_spike(si);
      if (retval) SETIERROR(retval,"from result_spike()");
    }
  }
  return 0;
}

